from pathlib import Path
from typing import Dict

from truss.model_framework import ModelFramework
from truss.model_inference import infer_sklearn_packages
from truss.templates.server.common.util import model_supports_predict_proba
from truss.types import ModelFrameworkType

MODEL_FILENAME = 'model.joblib'


class SKLearn(ModelFramework):

    def typ(self) -> ModelFrameworkType:
        return ModelFrameworkType.SKLEARN

    def infer_requirements(self) -> Dict[str, str]:
        return infer_sklearn_packages()

    def serialize_model_to_directory(self, model, target_directory: Path):
        import joblib
        model_filename = MODEL_FILENAME
        model_filepath = target_directory / model_filename
        joblib.dump(model, model_filepath, compress=True)

    def model_metadata(self, model) -> Dict[str, str]:
        supports_predict_proba = model_supports_predict_proba(model)
        return {
            'model_binary_dir': 'model',
            'supports_predict_proba': supports_predict_proba,
        }

    def supports_model_class(self, model_class) -> bool:
        model_framework, _, _ = model_class.__module__.partition('.')
        return model_framework == ModelFrameworkType.SKLEARN.value



import re

SECRET_NAME_MATCH_REGEX = re.compile(r'^[-._a-zA-Z0-9]+$')


def validate_secret_name(secret_name: str):
    if secret_name is None or not isinstance(secret_name, str) or secret_name == '':
        raise ValueError(f'Invalid secret name `{secret_name}`')

    def constraint_violation_msg():
        return f'Constraint violation for {secret_name}'

    if len(secret_name) > 253:
        raise ValueError(f'Secret name `{secret_name}` is longer than max allowed 253 chars.')

    if not SECRET_NAME_MATCH_REGEX.match(secret_name):
        raise ValueError(constraint_violation_msg() + ', invalid characters found in secret name.')

    if secret_name == '.':
        raise ValueError(constraint_violation_msg() + ', secret name cannot be `.`')

    if secret_name == '..':
        raise ValueError(constraint_violation_msg() + ', secret name cannot be `..`')



import tempfile
from pathlib import Path

from truss.contexts.local_loader.model_module_loader import \
    model_class_module_loaded

ORIG_MODEL_CLASS_CONTENT = '''
class Model:
    x = 1
'''

NEW_MODEL_CLASS_CONTENT = '''
class Model:
    x = 2
'''


UTIL_ACCESSOR_MODEL_FILE_CONTENT = '''
from model.util import X
from model.submodule.subfile import Y

class Model:
    x = X
    y = Y
'''

UTIL_FILE_CONTENT = '''
X = 1
'''

NEW_UTIL_FILE_CONTENT = '''
X = 2
'''

SUBMODULE_FILE_CONTENT = '''
Y = 3
'''

NEW_SUBMODULE_FILE_CONTENT = '''
Y = 4
'''


def test_model_module_finder():
    with tempfile.TemporaryDirectory() as temp_dir:
        _write_model_module_files(temp_dir, ORIG_MODEL_CLASS_CONTENT)
        with model_class_module_loaded(temp_dir, 'model.model') as model_module:
            model_class = getattr(model_module, 'Model')
            assert model_class.x == 1


def test_model_module_finder_reload():
    with tempfile.TemporaryDirectory() as temp_dir:
        _write_model_module_files(temp_dir, ORIG_MODEL_CLASS_CONTENT)
        with model_class_module_loaded(temp_dir, 'model.model') as model_module:
            model_class = getattr(model_module, 'Model')
            assert model_class.x == 1
    with tempfile.TemporaryDirectory() as temp_dir:
        _write_model_module_files(temp_dir, NEW_MODEL_CLASS_CONTENT)
        with model_class_module_loaded(temp_dir, 'model.model') as model_module:
            model_class = getattr(model_module, 'Model')
            assert model_class.x == 2


def test_model_module_finder_reload_non_model_file_updated():
    with tempfile.TemporaryDirectory() as temp_dir:
        _write_model_module_files(
            temp_dir,
            UTIL_ACCESSOR_MODEL_FILE_CONTENT,
            UTIL_FILE_CONTENT,
            SUBMODULE_FILE_CONTENT,
        )
        with model_class_module_loaded(temp_dir, 'model.model') as model_module:
            model_class = getattr(model_module, 'Model')
            assert model_class.x == 1
            assert model_class.y == 3
    with tempfile.TemporaryDirectory() as temp_dir:
        _write_model_module_files(
            temp_dir,
            UTIL_ACCESSOR_MODEL_FILE_CONTENT,
            NEW_UTIL_FILE_CONTENT,
            NEW_SUBMODULE_FILE_CONTENT,
        )
        with model_class_module_loaded(temp_dir, 'model.model') as model_module:
            model_class = getattr(model_module, 'Model')
            assert model_class.x == 2
            assert model_class.y == 4


def _write_model_module_files(
    scaffold_dir: str,
    model_file_content: str,
    util_file_content: str = None,
    submodule_file_content: str = None,
):
    model_dir_path = Path(scaffold_dir) / 'model'
    model_dir_path.mkdir()
    with (model_dir_path / 'model.py').open('w') as f:
        f.write(model_file_content)
    if util_file_content is not None:
        with (model_dir_path / 'util.py').open('w') as f:
            f.write(util_file_content)

    if submodule_file_content is not None:
        submodule_path = model_dir_path / 'submodule'
        submodule_path.mkdir()
        with (submodule_path / 'subfile.py').open('w') as f:
            f.write(submodule_file_content)



import importlib
import inspect
import logging
import traceback
from pathlib import Path
from typing import Dict, List

import kfserving
from secrets_resolver import SecretsResolver

MODEL_BASENAME = 'model'


class ModelWrapper(kfserving.KFModel):
    def __init__(self, config: dict):
        super().__init__(MODEL_BASENAME)
        self._config = config
        self._model = None

    def load(self):
        model_module_name = str(Path(self._config['model_class_filename']).with_suffix(''))
        module = importlib.import_module(f"{self._config['model_module_dir']}.{model_module_name}")
        model_class = getattr(module, self._config['model_class_name'])
        model_class_signature = inspect.signature(model_class)
        model_init_params = {}
        if _signature_accepts_keyword_arg(model_class_signature, 'config'):
            model_init_params['config'] = self._config
        if _signature_accepts_keyword_arg(model_class_signature, 'data_dir'):
            model_init_params['data_dir'] = Path('data')
        if _signature_accepts_keyword_arg(model_class_signature, 'secrets'):
            model_init_params['secrets'] = SecretsResolver.get_secrets(self._config)
        self._model = model_class(**model_init_params)
        if hasattr(self._model, 'load'):
            self._model.load()
        self.ready = True

    def preprocess(self, request: Dict) -> Dict:
        if not hasattr(self._model, 'preprocess'):
            return request
        return self._model.preprocess(request)

    def postprocess(self, request: Dict) -> Dict:
        if not hasattr(self._model, 'postprocess'):
            return request
        return self._model.postprocess(request)

    def predict(self, request: Dict) -> Dict[str, List]:
        response = {}
        try:
            return self._model.predict(request)
        except Exception:
            logging.error(traceback.format_exc())
            response['error'] = {'traceback': traceback.format_exc()}
            return response


def _signature_accepts_keyword_arg(signature: inspect.Signature, kwarg: str) -> bool:
    return kwarg in signature.parameters or _signature_accepts_kwargs(signature)


def _signature_accepts_kwargs(signature: inspect.Signature) -> bool:
    for param in signature.parameters.values():
        if param.kind == inspect.Parameter.VAR_KEYWORD:
            return True
    return False



from pathlib import Path
from typing import Dict

from truss.model_framework import ModelFramework
from truss.model_inference import infer_pytorch_packages
from truss.types import ModelFrameworkType

TORCH_PACKAGE_FILE = 'model_package.pt'
TORCH_MODEL_PICKLE_FILENAME = 'model.pkl'
TORCH_MODEL_PACKAGE_NAME = 'torch_model'


class PyTorch(ModelFramework):

    def typ(self) -> ModelFrameworkType:
        return ModelFrameworkType.PYTORCH

    def infer_requirements(self) -> Dict[str, str]:
        return infer_pytorch_packages()

    def serialize_model_to_directory(self, model, target_directory: Path):
        from torch import package
        try:
            _torch_package(model, target_directory / TORCH_PACKAGE_FILE, [])
        except package.package_exporter.PackagingError as pkg_err:
            # Make sure previous, potentially partially written, package is gone
            (target_directory / TORCH_PACKAGE_FILE).unlink()
            modules_to_extern = _broken_modules(pkg_err)
            _torch_package(model, target_directory / TORCH_PACKAGE_FILE, modules_to_extern)

    def model_metadata(self, model) -> Dict[str, str]:
        return {
            'torch_package_file': TORCH_PACKAGE_FILE,
            'torch_model_pickle_filename': TORCH_MODEL_PICKLE_FILENAME,
            'torch_model_package_name': TORCH_MODEL_PACKAGE_NAME,
            'model_binary_dir': 'model',
        }

    def supports_model_class(self, model_class) -> bool:
        try:
            import torch
            return issubclass(model_class, torch.nn.Module)
        except ModuleNotFoundError:
            return False


def _torch_package(model, path: Path, extern_modules: list):
    from torch import package
    with package.PackageExporter(path) as exp:
        exp.intern(f'{model.__class__.__module__}.**')
        for extern_module in extern_modules:
            exp.extern(f'{extern_module}.**')
        exp.save_pickle(TORCH_MODEL_PACKAGE_NAME, TORCH_MODEL_PICKLE_FILENAME, model)


def _broken_modules(pkg_error):
    """Extract the broken modules from the torch package error.

    We would extern these modules.
    Args:
        pkg_error (package.package_exporter.PackagingError): Error to extract broken modules from.
    """
    return [module_name
            for module_name, attrs in pkg_error.dependency_graph.nodes.items()
            if attrs.get('error') is not None]



import inspect
import os
import pathlib
import sys
from ast import ClassDef, FunctionDef
from dataclasses import dataclass
from typing import Any, Dict, List, Tuple

import pkg_resources
from packaging import version
from pkg_resources.extern.packaging.requirements import InvalidRequirement

from truss.constants import (HUGGINGFACE_TRANSFORMER, KERAS, PYTORCH, SKLEARN,
                             TENSORFLOW)
from truss.errors import FrameworkNotSupportedError

# list from https://scikit-learn.org/stable/developers/advanced_installation.html
SKLEARN_REQ_MODULE_NAME = {
    'numpy',
    'scipy',
    'joblib',
    'scikit-learn',
    'threadpoolctl',
}

# list from https://www.tensorflow.org/install/pip
# if problematic, lets look to https://www.tensorflow.org/install/source
TENSORFLOW_REQ_MODULE_NAME = {
    'tensorflow',
}


# list from https://pytorch.org/get-started/locally/
PYTORCH_REQ_MODULE_NAME = {
    'torch',
    'torchvision',
    'torchaudio',
}

HUGGINGFACE_TRANSFORMER_MODULE_NAME = {
    'transformers'
}

# lists of versions supported by the truss+base_images
PYTHON_VERSIONS = {
    'py37',
    'py38',
    'py39',
}


def pip_freeze():
    """
    This spawns a subprocess to do a pip freeze programmatically. pip is generally not supported as an API or threadsafe

    Returns: The result of a `pip freeze`

    """
    import pip
    pip_version = pip.__version__
    if version.parse(pip_version) < version.parse('20.1'):
        stream = os.popen('pip freeze -qq')
    else:
        stream = os.popen('pip list --format=freeze')
    this_env_requirements = [line.strip() for line in stream.readlines()]
    return this_env_requirements


def _get_entries_for_packages(list_of_requirements, desired_requirements):
    name_to_req_str = {}
    for req_name in desired_requirements:
        for full_req_str in list_of_requirements:
            if req_name == full_req_str.split('==')[0]:
                name_to_req_str[req_name] = full_req_str
    return name_to_req_str


def infer_sklearn_packages():
    return _get_entries_for_packages(pip_freeze(), SKLEARN_REQ_MODULE_NAME)


def infer_tensorflow_packages():
    return _get_entries_for_packages(pip_freeze(), TENSORFLOW_REQ_MODULE_NAME)


def infer_keras_packages():
    return _get_entries_for_packages(pip_freeze(), TENSORFLOW_REQ_MODULE_NAME)


def infer_pytorch_packages():
    return _get_entries_for_packages(pip_freeze(), PYTORCH_REQ_MODULE_NAME)


def infer_huggingface_packages():
    return _get_entries_for_packages(pip_freeze(), HUGGINGFACE_TRANSFORMER_MODULE_NAME)


def _infer_model_framework(model_class: str):
    model_framework, _, _ = model_class.__module__.partition('.')
    if model_framework == 'transformers':
        return HUGGINGFACE_TRANSFORMER
    if model_framework not in {SKLEARN, TENSORFLOW, KERAS}:
        try:
            import torch
            if issubclass(model_class, torch.nn.Module):
                model_framework = PYTORCH
            else:
                raise FrameworkNotSupportedError(f'Models must be one of '
                                                 f'{HUGGINGFACE_TRANSFORMER}, {SKLEARN}, {TENSORFLOW}, or {PYTORCH}.')
        except ModuleNotFoundError:
            raise FrameworkNotSupportedError(f'Models must be one of '
                                             f'{HUGGINGFACE_TRANSFORMER}, {SKLEARN}, {TENSORFLOW}, or {PYTORCH}.')

    return model_framework


@dataclass
class ModelBuildStageOne:
    # the Python Class of the model
    model_type: str
    # the framework that the model is built in
    model_framework: str


def _model_class(model: Any):
    return model.__class__


def infer_python_version() -> str:
    python_major_minor = f'py{sys.version_info.major}{sys.version_info.minor}'
    # might want to fix up this logic
    if python_major_minor not in PYTHON_VERSIONS:
        python_major_minor = None
    return python_major_minor


def infer_model_information(model: Any) -> ModelBuildStageOne:
    model_class = _model_class(model)
    model_framework = _infer_model_framework(model_class)
    model_type = model_class.__name__

    return ModelBuildStageOne(
        model_type,
        model_framework,
    )


def parse_requirements_file(requirements_file: str) -> dict:
    name_to_req_str = {}
    with pathlib.Path(requirements_file).open() as reqs_file:
        for raw_req in reqs_file.readlines():
            try:
                req = pkg_resources.Requirement.parse(raw_req)
                if req.specifier:
                    name_to_req_str[req.name] = str(req)
                else:
                    name_to_req_str[str(req)] = str(req)
            except InvalidRequirement:
                # there might be pip requirements that do not conform
                raw_req = str(raw_req).strip()
                name_to_req_str[f'custom_{raw_req}'] = raw_req
            except ValueError:
                # can't parse empty lines
                pass

    return name_to_req_str


def _infer_model_init_parameters(model_class: Any) -> Tuple[List, List]:
    full_arg_spec = inspect.getfullargspec(model_class.__init__)
    named_args = full_arg_spec.args[1:]
    number_of_kwargs = full_arg_spec.defaults and len(full_arg_spec.defaults) or 0
    required_args = full_arg_spec.args[1:-number_of_kwargs]
    return named_args, required_args


def _infer_model_init_parameters_ast(model_class_def: ClassDef) -> Tuple[List, List]:
    named_args = []
    required_args = []
    init_model_functions = [
        node for node in model_class_def.body if isinstance(node, FunctionDef) and node.name == '__init__'
    ]

    if not init_model_functions:
        return named_args, required_args

    assert len(init_model_functions) == 1, 'There should only be one __init__ function in the model class'
    init_model_function = init_model_functions[0]
    named_args = [arg.arg for arg in init_model_function.args.args][1:]
    number_of_defaults = len(init_model_function.args.defaults)
    required_args = named_args[:-number_of_defaults]
    return named_args, required_args


def validate_provided_parameters_with_model(model_class: Any, provided_parameters: Dict[str, Any]) -> None:
    """
    Validates that all provided parameters match the signature of the model.

    Args:
        model_class: The model class to validate against
        provided_parameters: The parameters to validate
    """
    if type(model_class) == ClassDef:
        named_args, required_args = _infer_model_init_parameters_ast(model_class)
    else:
        named_args, required_args = _infer_model_init_parameters(model_class)

    # Check that there are no extra parameters
    if not named_args:
        return

    if provided_parameters and type(provided_parameters) is not dict:
        raise TypeError(f'Provided parameters must be a dict, not {type(provided_parameters)}')

    for arg in provided_parameters:
        if arg not in named_args:
            raise ValueError(f'Provided parameter {arg} is not a valid init parameter for the model.')

    for arg in required_args:
        if arg not in provided_parameters:
            raise ValueError(f'Required init parameter {arg} was not provided for this model.')



from enum import Enum


class ModelFrameworkType(Enum):
    SKLEARN = 'sklearn'
    TENSORFLOW = 'tensorflow'
    KERAS = 'keras'
    PYTORCH = 'pytorch'
    HUGGINGFACE_TRANSFORMER = 'huggingface_transformer'
    CUSTOM = 'custom'



import logging
from pathlib import Path
from typing import Dict

from truss.model_framework import ModelFramework
from truss.model_inference import infer_huggingface_packages
from truss.types import ModelFrameworkType

logger = logging.getLogger(__name__)


class HuggingfaceTransformer(ModelFramework):

    def typ(self) -> ModelFrameworkType:
        return ModelFrameworkType.HUGGINGFACE_TRANSFORMER

    def infer_requirements(self) -> Dict[str, str]:
        return infer_huggingface_packages()

    def serialize_model_to_directory(self, model, target_directory: Path):
        # For Huggingface models, all the important details are in metadata.
        pass

    def model_metadata(self, model) -> Dict[str, str]:
        hf_task = self.model_type(model)
        return {
            'transformer_config': {
                'model': _hf_model_name(model),
            },
            'has_hybrid_args': hf_task in {'text-generation'},
            'has_named_args': hf_task in {'zero-shot-classification'},
        }

    def model_type(self, model) -> str:
        return _infer_hf_task(model)

    def supports_model_class(self, model_class) -> bool:
        model_framework, _, _ = model_class.__module__.partition('.')
        return model_framework == 'transformers'


def _hf_model_name(model) -> str:
    try:
        return model.model.config._name_or_path
    except AttributeError:
        logger.info('Unable to infer a HuggingFace model on this task pipeline; transformer library will default')


def _infer_hf_task(model) -> str:
    try:
        return model.task
    except AttributeError as error:
        logger.exception('Unable to find a HuggingFace task on this object, did you call with a `Pipeline` object')
        raise error



from pathlib import Path

from jinja2 import Template

from truss.constants import (MODEL_DOCKERFILE_NAME, REQUIREMENTS_TXT_FILENAME,
                             SERVER_CODE_DIR, SERVER_DOCKERFILE_TEMPLATE_NAME,
                             SERVER_REQUIREMENTS_TXT_FILENAME,
                             SYSTEM_PACKAGES_TXT_FILENAME, TEMPLATES_DIR)
from truss.contexts.truss_context import TrussContext
from truss.docker import Docker
from truss.truss_spec import TrussSpec
from truss.utils import (build_truss_target_directory, copy_file_path,
                         copy_tree_path, given_or_temporary_dir)

BUILD_SERVER_DIR_NAME = 'server'


class ImageBuilderContext(TrussContext):
    @staticmethod
    def run(truss_dir: Path):
        return ImageBuilder(truss_dir)


class ImageBuilder:
    def __init__(self, truss_dir: Path) -> None:
        self._spec = TrussSpec(truss_dir)

    def build_image(self, build_dir: Path = None, tag: str = None, labels: dict = None):
        """Build image.

        Arguments:
            build_dir(Path): Directory to use for building the docker image. If None
                             then a temporary directory is used.
            tag(str): A tag to assign to the docker image.
        """

        with given_or_temporary_dir(build_dir) as build_dir_path:
            self.prepare_image_build_dir(build_dir_path)
            return Docker.client().build(
                str(build_dir_path),
                labels=labels if labels else {},
                tags=tag or self.default_tag)

    @property
    def default_tag(self):
        return f'{self._spec.model_framework_name}-model:latest'

    def prepare_image_build_dir(self, build_dir: Path = None):
        """Prepare a directory for building the docker image from.

        Returns:
            docker command to build the docker image.
        """
        if build_dir is None:
            build_dir = build_truss_target_directory(self._spec.model_framework_name)
            # todo: Add a logging statement here, suggesting how to clean up the directory.

        copy_tree_path(self._spec.truss_dir, build_dir)
        copy_tree_path(
            SERVER_CODE_DIR,
            build_dir / BUILD_SERVER_DIR_NAME,
        )
        copy_file_path(
            TEMPLATES_DIR / self._spec.model_framework_name / REQUIREMENTS_TXT_FILENAME,
            build_dir / SERVER_REQUIREMENTS_TXT_FILENAME,
        )

        with (build_dir / REQUIREMENTS_TXT_FILENAME).open('w') as req_file:
            req_file.write(self._spec.requirements_txt)

        with (build_dir / SYSTEM_PACKAGES_TXT_FILENAME).open('w') as req_file:
            req_file.write(self._spec.system_packages_txt)

        dockerfile_template_path = TEMPLATES_DIR / SERVER_DOCKERFILE_TEMPLATE_NAME

        with dockerfile_template_path.open() as dockerfile_template_file:
            dockerfile_template = Template(dockerfile_template_file.read())
            dockerfile_contents = dockerfile_template.render(config=self._spec.config)
            docker_file_path = build_dir / MODEL_DOCKERFILE_NAME
            with docker_file_path.open('w') as docker_file:
                docker_file.write(dockerfile_contents)

    def docker_build_command(self, build_dir) -> str:
        return f'docker build {build_dir} -t {self.default_tag}'



import json
import logging
import os
from functools import wraps
from pathlib import Path

import click

import truss

logging.basicConfig(level=logging.INFO)


def echo_output(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        click.echo(f(*args, **kwargs))
    return wrapper


def error_handling(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        try:
            f(*args, **kwargs)
        except ValueError as e:
            click.echo(e)
            print_help()
        except Exception as e:
            click.echo(e)
    return wrapper


def print_help():
    ctx = click.get_current_context()
    click.echo(ctx.get_help())


@click.group(name="truss")
def cli_group():
    pass


@cli_group.command()
@click.argument("target_directory", required=True)
@error_handling
def init(target_directory):
    """ Initializes an empty Truss directory.

    TARGET_DIRECTORY: A Truss is created in this directory
    """
    tr_path = Path(target_directory)
    if click.confirm(f'A Truss will be created at {tr_path}'):
        truss.init(target_directory=target_directory)
        click.echo(f"Truss was created in {tr_path}")


@cli_group.command()
@click.argument("build_dir")
@click.argument("target_directory", required=False)
@error_handling
def build_context(build_dir, target_directory) -> None:
    """
    Create a docker build context for a Truss.

    BUILD_DIR: Folder where image context is built for Truss

    TARGET DIRECTORY: A Truss directory. If none, use current directory.
    """
    tr = _get_truss_from_directory(target_directory=target_directory)
    tr.docker_build_setup(build_dir=Path(build_dir))


@cli_group.command()
@click.argument("target_directory", required=False)
@click.argument("build_dir", required=False)
@error_handling
@click.option("--tag", help="Docker image tag")
def build_image(target_directory, build_dir, tag):
    """
    Builds the docker image for a Truss.

    TARGET DIRECTORY: A Truss directory. If none, use current directory.

    BUILD_DIR: Image context. If none, a temp directory is created.
    """
    tr = _get_truss_from_directory(target_directory=target_directory)
    if build_dir:
        build_dir = Path(build_dir)
    tr.build_docker_image(build_dir=build_dir, tag=tag)


@cli_group.command()
@click.argument("target_directory", required=False)
@click.argument("build_dir", required=False)
@click.option("--tag", help="Docker build image tag")
@click.option("--port", type=int, default=8080, help="Local port used to run image")
@click.option(
    "--detach", is_flag=True, default=False, help="Flag for detaching process"
)
@error_handling
def run_image(target_directory, build_dir, tag, port, detach):
    """
    Runs the docker image for a Truss.

    TARGET DIRECTORY: A Truss directory. If none, use current directory.

    BUILD_DIR: Image context. If none, a temp directory is created.
    """
    tr = _get_truss_from_directory(target_directory=target_directory)
    if build_dir:
        build_dir = Path(build_dir)
    urls = tr.get_urls_from_truss()
    if urls:
        click.confirm(f"Container already exists at {urls}. Are you sure you want to continue?")
    tr.docker_run(build_dir=build_dir, tag=tag, local_port=port, detach=detach)


@cli_group.command()
@click.option("--target_directory", required=False, help="Directory of truss")
@click.option("--request", type=str, required=False, help="String formatted as json that represents request")
@click.option("--build-dir", type=click.Path(exists=True), required=False, help="Directory where context is built")
@click.option("--tag", help="Docker build image tag")
@click.option("--port", type=int, default=8080, help="Local port used to run image")
@click.option("--run-local", is_flag=True, default=False, help="Flag to run prediction locally")
@click.option("--request-file", type=click.Path(exists=True), help="Path to json file containing the request")
@error_handling
@echo_output
def predict(target_directory, request, build_dir, tag, port, run_local, request_file):
    """Runs prediction for a Truss in a docker image or locally """
    if request is not None:
        request_data = json.loads(request)
    elif request_file is not None:
        with open(request_file) as json_file:
            request_data = json.load(json_file)
    else:
        raise ValueError("At least one of request or request-file must be supplied.")

    tr = _get_truss_from_directory(target_directory=target_directory)
    if run_local:
        return tr.server_predict(request_data)
    else:
        return tr.docker_predict(
            request_data, build_dir=build_dir,
            tag=tag, local_port=port, detach=True
        )


@cli_group.command()
@click.argument("target_directory", required=False)
@click.option("--name", type=str, required=False, help="Name of example to run")
@click.option(
    "--local", is_flag=True, default=False, help="Flag to run prediction locally"
)
@error_handling
@echo_output
def run_example(target_directory, name, local):
    """
    Runs examples specified in the Truss, over docker.

    TARGET DIRECTORY: A Truss directory. If none, use current directory.
    """
    tr = _get_truss_from_directory(target_directory=target_directory)
    predict_fn = tr.docker_predict
    if local:
        predict_fn = tr.server_predict

    if name is not None:
        example = tr.example(name)
        click.echo(f'Running example: {name}')
        return predict_fn(example)
    else:
        example_outputs = []
        for name, example in tr.examples().items():
            click.echo(f'Running example: {name}')
            example_outputs.append(predict_fn(example))
        return example_outputs


@cli_group.command()
@click.argument("target_directory", required=False)
@error_handling
def get_container_logs(target_directory):
    """
    Get logs in a container is running for a truss

    TARGET DIRECTORY: A Truss directory. If none, use current directory.
    """
    for log in _get_truss_from_directory(
        target_directory=target_directory
    ).container_logs():
        click.echo(log)


@cli_group.command()
@click.argument("target_directory", required=False)
def kill(target_directory):
    """
    Kills containers related to truss.

    TARGET DIRECTORY: A Truss directory. If none, use current directory.
    """
    tr = _get_truss_from_directory(target_directory=target_directory)
    tr.kill_container()


@cli_group.command()
def kill_all():
    "Kills all truss containers that are not manually persisted"
    truss.kill_all()


def _get_truss_from_directory(target_directory: str = None):
    """Gets Truss from directory. If none, use the current directory"""
    if target_directory is None:
        target_directory = os.getcwd()
    return truss.from_directory(target_directory)


if __name__ == "__main__":
    cli_group()



import importlib
import sys
from contextlib import contextmanager
from importlib.machinery import PathFinder


class ModelModuleFinder(PathFinder):
    _truss_dir: str

    @classmethod
    def set_model_truss_dir(cls, truss_dir: str):
        cls._truss_dir = truss_dir
        cls.add_to_meta_path()

    @classmethod
    def find_spec(cls, fullname, path=None, target=None):
        if not path:
            path = [cls._truss_dir]
        else:
            path = list(path) + [cls._truss_dir]
        return PathFinder.find_spec(fullname, path, target)

    @classmethod
    def add_to_meta_path(cls):
        if cls not in sys.meta_path:
            sys.meta_path.append(cls)

    @classmethod
    def remove_from_meta_path(cls):
        if cls in sys.meta_path:
            del sys.meta_path[sys.meta_path.index(cls)]


class ModelModuleLoader:
    @staticmethod
    def import_model_module(name):
        """
        Custom import implementation for loading model module.

        Based on https://docs.python.org/3/library/importlib.html#approximating-importlib-import-module
        """
        absolute_name = importlib.util.resolve_name(name, None)
        path = None
        if '.' in absolute_name:
            parent_name, _, child_name = absolute_name.rpartition('.')
            parent_module = ModelModuleLoader.import_model_module(parent_name)
            path = parent_module.__spec__.submodule_search_locations

        spec = ModelModuleFinder.find_spec(absolute_name, path)
        module = importlib.util.module_from_spec(spec)
        sys.modules[absolute_name] = module
        spec.loader.exec_module(module)
        if path is not None:
            setattr(parent_module, child_name, module)
        return module


@contextmanager
def model_class_module_loaded(truss_dir: str, model_class_module_fullname: str):
    try:
        model_module_name = model_class_module_fullname.split('.')[0]
        # Unload so that all model modules can be freshly loaded.
        _unload_model_modules(model_module_name)
        ModelModuleFinder.set_model_truss_dir(truss_dir)
        model_class_module = ModelModuleLoader.import_model_module(model_class_module_fullname)
        yield model_class_module
    finally:
        ModelModuleFinder.remove_from_meta_path()


def _unload_model_modules(model_module_name: str):
    model_module_submodules = [module_name
                               for module_name in sys.modules
                               if module_name.startswith(model_module_name + '.')]
    for model_module_submodule in model_module_submodules:
        del sys.modules[model_module_submodule]



import logging
from typing import Dict

from truss.constants import TRUSS_DIR


class Docker:
    _client = None

    @staticmethod
    def client():
        if Docker._client is None:
            from python_on_whales import docker
            Docker._client = docker
        return Docker._client


def get_containers(labels: dict, all=False):
    """Gets containers given labels."""
    return Docker.client().container.list(filters=_create_label_filters(labels), all=all)


def get_images(labels: dict):
    """Gets images given labels."""
    return Docker.client().image.list(filters=_create_label_filters(labels))


def get_urls_from_container(container):
    """Gets url where docker container is hosted."""
    ports = container.network_settings.ports
    valid_urls = []
    for port in ports.values():
        for urls in port:
            valid_urls.append("http://" + urls["HostIp"] + ":" + urls["HostPort"])
    return valid_urls


def kill_containers(labels: Dict[str, str]):
    containers = get_containers(labels)
    for container in containers:
        container_labels = container.config.labels
        if TRUSS_DIR in container_labels:
            truss_dir = container_labels[TRUSS_DIR]
            logging.info(f"Killing Container: {container.id} for {truss_dir}")
    Docker.client().container.kill(containers)


def get_container_logs(container):
    return Docker.client().container.logs(container, follow=True, stream=True)


def _create_label_filters(labels: dict):
    return {
        f'label={label_key}': label_value
        for label_key, label_value in labels.items()
    }



from unittest import mock

from truss.templates.server.common.util import \
    assign_request_to_inputs_instances_after_validation


def test_assign_request_to_inputs_instances_after_validation():
    inputs_input = [1, 2, 3, 4]
    inputs_dict = {'inputs': inputs_input}
    instances_input = [5, 6, 7, 8]
    instances_dict = {'instances': instances_input}

    processed_inputs = assign_request_to_inputs_instances_after_validation(inputs_dict)
    processed_instances = assign_request_to_inputs_instances_after_validation(instances_dict)

    assert processed_inputs['instances'] == processed_inputs['inputs'] == inputs_input
    assert processed_instances['instances'] == processed_instances['inputs'] == instances_input


def model_supports_predict_proba():
    mock_not_predict_proba = mock.Mock(name='mock_not_predict_proba')
    mock_not_predict_proba.predict_proba.return_value = False

    mock_check_proba = mock.Mock(name='mock_check_proba')
    mock_check_proba.predict_proba.return_value = True
    mock_check_proba._check_proba.return_value = True

    mock_not_check_proba = mock.Mock(name='mock_not_check_proba')
    mock_not_check_proba.predict_proba.return_value = True
    mock_not_check_proba._check_proba.side_effect = AttributeError

    assert not model_supports_predict_proba(mock_not_predict_proba)
    assert model_supports_predict_proba(mock_check_proba)
    assert not model_supports_predict_proba(mock_not_check_proba)



try:
    import kfserving  # noqa: F401
    KFSERVING_LIB = True
except ModuleNotFoundError:
    KFSERVING_LIB = False


def ensure_kfserving_installed():
    if not KFSERVING_LIB:
        raise ModuleNotFoundError('Could not successfully import "kfserving" package, check your Python environment')
    return True



from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict

import yaml


@dataclass
class LocalConfig:
    secrets: Dict[str, str] = field(default_factory=dict)

    @staticmethod
    def from_dict(d):
        return LocalConfig(
            secrets=d.get('secrets', {}),
        )

    @staticmethod
    def from_yaml(yaml_path: Path):
        with yaml_path.open() as yaml_file:
            return LocalConfig.from_dict(yaml.safe_load(yaml_file))

    def to_dict(self):
        return {
            'secrets': self.secrets,
        }

    def write_to_yaml_file(self, path: Path):
        with path.open('w') as config_file:
            yaml.dump(self.to_dict(), config_file)



import pytest

from truss.constants import PYTORCH
from truss.model_inference import (infer_model_information,
                                   validate_provided_parameters_with_model)


def test_pytorch_init_arg_validation(pytorch_model_with_init_args, pytorch_model_init_args):
    pytorch_model_with_init_args, _ = pytorch_model_with_init_args
    # Validates with args and kwargs
    validate_provided_parameters_with_model(pytorch_model_with_init_args.__class__, pytorch_model_init_args)

    # Errors if bad args
    with pytest.raises(ValueError):
        validate_provided_parameters_with_model(pytorch_model_with_init_args.__class__, {'foo': 'bar'})

    # Validates with only args
    copied_args = pytorch_model_init_args.copy()
    copied_args.pop('kwarg1')
    copied_args.pop('kwarg2')
    validate_provided_parameters_with_model(pytorch_model_with_init_args, copied_args)

    # Requires all args
    with pytest.raises(ValueError):
        validate_provided_parameters_with_model(pytorch_model_with_init_args, {})


def test_infer_model_information(pytorch_model_with_init_args):
    model_info = infer_model_information(pytorch_model_with_init_args[0])
    assert model_info.model_framework == PYTORCH
    assert model_info.model_type == 'MyModel'



from pathlib import Path
from typing import Dict

from truss.model_framework import ModelFramework
from truss.model_inference import infer_keras_packages
from truss.types import ModelFrameworkType


class Keras(ModelFramework):

    def typ(self) -> ModelFrameworkType:
        return ModelFrameworkType.KERAS

    def infer_requirements(self) -> Dict[str, str]:
        return infer_keras_packages()

    def serialize_model_to_directory(self, model, target_directory: Path):
        model.save(target_directory)

    def model_metadata(self, model) -> Dict[str, str]:
        return {
            'model_binary_dir': 'model',
        }

    def supports_model_class(self, model_class) -> bool:
        model_framework, _, _ = model_class.__module__.partition('.')
        return model_framework in [ModelFrameworkType.KERAS.value, ModelFrameworkType.TENSORFLOW.value]



import os
import time

from truss import utils


def test_max_modified():
    epoch_time = int(time.time())
    with utils.given_or_temporary_dir() as dir:
        t1 = utils.get_max_modified_time_of_dir(dir)
        assert(t1 > epoch_time)
        time.sleep(0.1)
        os.makedirs(os.path.join(dir, "test"))
        t2 = utils.get_max_modified_time_of_dir(dir)
        assert(t2 > t1)


test_max_modified()



import os
import random
import string
import tempfile
from contextlib import contextmanager
from distutils.dir_util import copy_tree
from distutils.file_util import copy_file
from pathlib import Path


def copy_tree_path(src: Path, dest: Path):
    return copy_tree(str(src), str(dest))


def copy_file_path(src: Path, dest: Path):
    return copy_file(str(src), str(dest))


def get_max_modified_time_of_dir(path: Path):
    max_modified_time = os.path.getmtime(path)
    for root, dirs, files in os.walk(path):
        if os.path.islink(root):
            raise ValueError(f'Symlinks not allowed in Truss: {root}')
        files = [f for f in files if not f.startswith('.')]
        dirs[:] = [d for d in dirs if not d.startswith('.')]
        max_modified_time = max(max_modified_time, os.path.getmtime(root))
        for file in files:
            max_modified_time = max(
                max_modified_time,
                os.path.getmtime(os.path.join(root, file))
            )
    return max_modified_time


@contextmanager
def given_or_temporary_dir(given_dir: Path = None):
    if given_dir is not None:
        yield given_dir
    else:
        with tempfile.TemporaryDirectory() as temp_dir:
            yield Path(temp_dir)


def build_truss_target_directory(model_framework_name: str) -> Path:
    """Builds a directory under ~/.truss/models for the purpose of creating a Truss at."""
    rand_suffix = ''.join(random.choices(string.ascii_uppercase + string.digits, k=6))
    target_directory_path = Path(
        Path.home(),
        '.truss',
        'models',
        f'{model_framework_name}-{rand_suffix}'

    )
    target_directory_path.mkdir(parents=True)
    return target_directory_path



import time

import pytest

from truss.build import kill_all
from truss.constants import TRUSS
from truss.docker import Docker, get_containers
from truss.local.local_config_handler import LocalConfigHandler
from truss.truss_handle import TrussHandle


def test_spec(custom_model_truss_dir_with_pre_and_post):
    dir_path = custom_model_truss_dir_with_pre_and_post
    sc = TrussHandle(dir_path)
    spec = sc.spec
    assert spec.truss_dir == dir_path


def test_server_predict(custom_model_truss_dir_with_pre_and_post):
    sc = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    resp = sc.server_predict({
        'inputs': [1, 2, 3, 4],
    })
    assert resp == {'predictions': [4, 5, 6, 7]}


@pytest.mark.integration
def test_build_docker_image(custom_model_truss_dir_with_pre_and_post):
    sc = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    tag = 'test-build-image-tag:0.0.1'
    image = sc.build_docker_image(tag=tag)
    assert image.repo_tags[0] == tag


@pytest.mark.integration
def test_build_docker_image_gpu(custom_model_truss_dir_for_gpu, tmp_path):
    sc = TrussHandle(custom_model_truss_dir_for_gpu)
    tag = 'test-build-image-gpu-tag:0.0.1'
    build_dir = tmp_path / 'scaffold_build_dir'
    image = sc.build_docker_image(tag=tag, build_dir=build_dir)
    assert image.repo_tags[0] == tag


@pytest.mark.integration
def test_docker_run(custom_model_truss_dir_with_pre_and_post):
    sc = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    tag = 'test-docker-run-tag:0.0.1'
    container = sc.docker_run(tag=tag)
    try:
        assert _container_exists(container)
    finally:
        Docker.client().kill(container)


@pytest.mark.integration
def test_docker_run_gpu(custom_model_truss_dir_for_gpu):
    sc = TrussHandle(custom_model_truss_dir_for_gpu)
    tag = 'test-docker-run-gpu-tag:0.0.1'
    container = sc.docker_run(tag=tag)
    try:
        assert _container_exists(container)
    finally:
        Docker.client().kill(container)


@pytest.mark.integration
def test_docker_run_without_tag(custom_model_truss_dir_with_pre_and_post):
    sc = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    container = sc.docker_run()
    try:
        assert _container_exists(container)
    finally:
        Docker.client().kill(container)


@pytest.mark.integration
def get_docker_containers_from_labels(
    custom_model_truss_dir_with_pre_and_post
):
    try:
        t1 = TrussHandle(custom_model_truss_dir_with_pre_and_post)
        assert(len(t1.get_docker_containers_from_labels()) == 0)
        t1.docker_run()
        assert(len(t1.get_docker_containers_from_labels()) == 1)
        t1.docker_run(port=3000)
        assert(len(t1.get_docker_containers_from_labels()) == 2)
        t1.kill_container()
        assert(len(t1.get_docker_containers_from_labels()) == 0)
    finally:
        _ensure_kill_all()


@pytest.mark.integration
def test_docker_predict(custom_model_truss_dir_with_pre_and_post):
    sc = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    tag = 'test-docker-predict-tag:0.0.1'
    try:
        result = sc.docker_predict({'inputs': [1, 2]}, tag=tag)
        assert result == {'predictions': [4, 5]}
    finally:
        _ensure_kill_all()


@pytest.mark.integration
def test_docker_multiple_predict(custom_model_truss_dir_with_pre_and_post):
    sc = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    tag = 'test-docker-predict-tag:0.0.1'
    try:
        r1 = sc.docker_predict({'inputs': [1, 2]}, tag=tag)
        r2 = sc.docker_predict({'inputs': [3, 4]}, tag=tag)
        assert r1 == {'predictions': [4, 5]}
        assert r2 == {'predictions': [6, 7]}
        assert(len(sc.get_docker_containers_from_labels()) == 1)
    finally:
        _ensure_kill_all()


@pytest.mark.integration
def test_kill_all(custom_model_truss_dir, custom_model_truss_dir_with_pre_and_post):
    t1 = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    t2 = TrussHandle(custom_model_truss_dir)
    try:
        t1.docker_run()
        assert(len(t1.get_docker_containers_from_labels()) == 1)
        t2.docker_run(local_port=3000)
        assert(len(t2.get_docker_containers_from_labels()) == 1)
        _ensure_kill_all()
        assert(len(t1.get_docker_containers_from_labels()) == 0)
        assert(len(t2.get_docker_containers_from_labels()) == 0)
    finally:
        _ensure_kill_all()


@pytest.mark.integration
def test_docker_predict_gpu(custom_model_truss_dir_for_gpu):
    sc = TrussHandle(custom_model_truss_dir_for_gpu)
    tag = 'test-docker-predict-gpu-tag:0.0.1'
    try:
        result = sc.docker_predict({'inputs': [1]}, tag=tag)
        assert result['predictions'][0]['cuda_version'].startswith('11')
    finally:
        _ensure_kill_all()


@pytest.mark.integration
def test_docker_predict_secrets(custom_model_truss_dir_for_secrets):
    sc = TrussHandle(custom_model_truss_dir_for_secrets)
    tag = 'test-docker-predict-secrets-tag:0.0.1'
    LocalConfigHandler.set_secret('secret_name', 'secret_value')
    try:
        result = sc.docker_predict({'inputs': ['secret_name']}, tag=tag)
        assert result['predictions'][0] == 'secret_value'
    finally:
        LocalConfigHandler.remove_secret('secret_name')
        _ensure_kill_all()


@pytest.mark.integration
def test_custom_python_requirement(custom_model_truss_dir_with_pre_and_post):
    sc = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    sc.add_python_requirement('theano')
    sc.add_python_requirement('scipy')
    tag = 'test-custom-python-req-tag:0.0.1'
    container = sc.docker_run(tag=tag)
    try:
        _verify_python_requirement_installed_on_container(tag, 'theano')
        _verify_python_requirement_installed_on_container(tag, 'scipy')
    finally:
        Docker.client().kill(container)


@pytest.mark.integration
def test_custom_system_package(custom_model_truss_dir_with_pre_and_post):
    sc = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    sc.add_system_package('jq')
    sc.add_system_package('fzf')
    tag = 'test-custom-system-package-tag:0.0.1'
    container = sc.docker_run(tag=tag)
    try:
        _verify_system_package_installed_on_container(tag, 'jq')
        _verify_system_package_installed_on_container(tag, 'fzf')
    finally:
        Docker.client().kill(container)


def test_enable_gpu(custom_model_truss_dir_with_pre_and_post):
    sc = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    sc.enable_gpu()
    assert sc.spec.config.resources.use_gpu


def test_update_requirements(custom_model_truss_dir_with_pre_and_post):
    sc = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    requirements = [
        'tensorflow==2.3.1',
        'uvicorn==0.12.2',
    ]
    sc.update_requirements(requirements)
    sc_requirements = sc.spec.requirements
    assert sc_requirements == requirements


def test_update_requirements_from_file(custom_model_truss_dir_with_pre_and_post, tmp_path):
    sc = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    requirements = [
        'tensorflow==2.3.1',
        'uvicorn==0.12.2',
    ]
    req_file_path = tmp_path / 'requirements.txt'
    with req_file_path.open('w') as req_file:
        for req in requirements:
            req_file.write(f'{req}\n')
    sc.update_requirements_from_file(str(req_file_path))
    sc_requirements = sc.spec.requirements
    assert sc_requirements == requirements


@pytest.mark.integration
def test_add_environment_variable(custom_model_truss_dir_with_pre_and_post):
    sc = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    sc.add_environment_variable('test_env', 'test_value')
    tag = 'test-add-env-var-tag:0.0.1'
    container = sc.docker_run(tag=tag)
    try:
        _verify_environment_variable_on_container(tag, 'test_env', 'test_value')
    finally:
        Docker.client().kill(container)


def test_add_data_file(custom_model_truss_dir_with_pre_and_post, tmp_path):
    sc = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    data_filepath = tmp_path / 'test_data.txt'
    with data_filepath.open('w') as data_file:
        data_file.write('test')

    sc.add_data(str(data_filepath))

    scaf_data_filepath = sc.spec.data_dir / 'test_data.txt'
    assert scaf_data_filepath.exists()
    with scaf_data_filepath.open() as data_file:
        assert data_file.read() == 'test'


def test_add_data_fileglob(custom_model_truss_dir_with_pre_and_post, tmp_path):
    sc = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    file1_path = tmp_path / 'test_data1.txt'
    with file1_path.open('w') as data_file:
        data_file.write('test1')

    file2_path = tmp_path / 'test_data2.txt'
    with file2_path.open('w') as data_file:
        data_file.write('test2')

    file2_path = tmp_path / 'test_data3.json'
    with file2_path.open('w') as data_file:
        data_file.write('{}')

    sc.add_data(f'{str(tmp_path)}/*.txt')

    assert (sc.spec.data_dir / 'test_data1.txt').exists()
    assert (sc.spec.data_dir / 'test_data2.txt').exists()
    assert not (sc.spec.data_dir / 'test_data2.json').exists()


def test_add_data_dir(custom_model_truss_dir_with_pre_and_post, tmp_path):
    sc = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    sub_dir = tmp_path / 'sub'
    sub_sub_dir = sub_dir / 'sub'
    sub_sub_dir.mkdir(parents=True)

    file_path = sub_sub_dir / 'test_file.txt'
    with file_path.open('w') as data_file:
        data_file.write('test')

    sc.add_data(str(sub_dir))

    scaf_file_path = sc.spec.data_dir / 'sub' / 'sub' / 'test_file.txt'
    assert scaf_file_path.exists()
    with scaf_file_path.open() as data_file:
        assert data_file.read() == 'test'


def test_examples(custom_model_truss_dir_with_pre_and_post):
    sc = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    examples = sc.examples()
    assert 'example1' in examples


def test_example(custom_model_truss_dir_with_pre_and_post):
    sc = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    assert 'inputs' in sc.example('example1')


def test_example_index(custom_model_truss_dir_with_pre_and_post):
    sc = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    assert 'inputs' in sc.example(0)


def test_add_example_new(custom_model_truss_dir_with_pre_and_post):
    sc = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    orig_examples = sc.examples()
    sc.add_example('example2', {'inputs': [[1]]})
    assert sc.examples() == {
        **orig_examples,
        'example2': {'inputs': [[1]]},
    }


def test_add_example_update(custom_model_truss_dir_with_pre_and_post):
    sc = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    sc.add_example('example1', {'inputs': [[1]]})
    assert sc.examples() == {
        'example1': {'inputs': [[1]]},
    }


def test_model_without_pre_post(custom_model_truss_dir):
    sc = TrussHandle(custom_model_truss_dir)
    resp = sc.server_predict({
        'inputs': [1, 2, 3, 4],
    })
    assert resp == [1, 1, 1, 1]


@pytest.mark.integration
def test_docker_predict_model_without_pre_post(custom_model_truss_dir):
    sc = TrussHandle(custom_model_truss_dir)
    try:
        resp = sc.docker_predict({
            'inputs': [1, 2, 3, 4],
        })
        assert resp == [1, 1, 1, 1]
    finally:
        _ensure_kill_all()


def _container_exists(container) -> bool:
    for row in Docker.client().ps():
        if row.id.startswith(container.id):
            return True
    return False


def _verify_system_package_installed_on_container(tag: str, cmd: str):
    resp = Docker.client().run(tag, ['which', cmd])
    assert resp.strip() == f'/usr/bin/{cmd}'


def _verify_python_requirement_installed_on_container(tag: str, cmd: str):
    resp = Docker.client().run(tag, ['pip', 'show', cmd])
    assert resp.splitlines()[0].lower() == f'Name: {cmd}'.lower()


def _verify_environment_variable_on_container(
    tag: str,
    env_var_name: str,
    env_var_value: str,
):
    resp = Docker.client().run(tag, ['env'])
    needle = f'{env_var_name}={env_var_value}'
    assert needle in resp.splitlines()


def _ensure_kill_all():
    kill_all()
    attempts = 0
    while attempts < 10:
        containers = get_containers({
            TRUSS: True
        })
        if len(containers) == 0:
            return
        attempts += 1
        time.sleep(1)



from truss.contexts.local_loader.load_local import _prepare_secrets
from truss.local.local_config_handler import LocalConfigHandler
from truss.truss_handle import TrussHandle
from truss.truss_spec import TrussSpec


def test_prepare_secrets(custom_model_truss_dir, tmp_path):
    orig_truss_config_dir = LocalConfigHandler.TRUSS_CONFIG_DIR
    LocalConfigHandler.TRUSS_CONFIG_DIR = tmp_path
    try:
        LocalConfigHandler.set_secret('secret_name', 'secret_value')
        handle = TrussHandle(custom_model_truss_dir)
        handle.add_secret('secret_name')
        spec = TrussSpec(custom_model_truss_dir)
        secrets = _prepare_secrets(spec)
        assert secrets['secret_name'] == 'secret_value'
    finally:
        LocalConfigHandler.TRUSS_CONFIG_DIR = orig_truss_config_dir
